# -*- coding: utf-8 -*-
#
# Copyright (C) 2018 Shinichi Takii, shinichi.takii@gmail.com
#
# This module is part of python-ddlparse and is released under
# the BSD License: https://opensource.org/licenses/BSD-3-Clause

from .ddlparse import *

__copyright__    = 'Copyright (C) 2018-2020 Shinichi Takii'
__version__      = '1.5.0'
__license__      = 'BSD-3-Clause'
__author__       = 'Shinichi Takii'
__author_email__ = 'shinichi.takii@gmail.com'
__url__          = 'http://github.com/shinichi-takii/ddlparse'

__all__ = ['DdlParse', 'DdlParseTable', 'DdlParseColumn', 'DdlParseColumnDict']
